--
-- (C) Miguel Oliveira e Silva (mos@det.ua.pt), October 2000
--

class MAIN

inherit
  THREAD_CONTROL

creation
  make

feature

  num_producer_stages: INTEGER is 3;

  num_consumers: INTEGER is 7;

  num_items: INTEGER is 100;

  make is
    local
      barr_init,barr_finish: THREAD_BARRIER;
      data: SHARED_DATA;
      mtx: MUTEX;
      cnd_var: CONDITION_VARIABLE;
      p: PRODUCER;
      c: CONSUMER;
      i: INTEGER;
      msg: STRING;
    do
      !!msg.make(80);
      !!data.make;
      !!mtx.make;
      !!cnd_var.make;
      !!barr_init.make_static(1+num_consumers+1); -- main+consumers+1producer
      !!barr_finish.make_static(1+num_consumers+1); -- main+consumers+1producer
      from
        i := 1
      until
        i > num_consumers
      loop
        msg.copy("Creating consumer ");
        msg.append(i.to_string);
        msg.append("...");
        print_line(msg);
        !!c.make(i,data,mtx,cnd_var,barr_init,barr_finish);
        c.start;
        i := i + 1;
      end;
      from
        i := 1
      until
        i > num_producer_stages
      loop
        msg.copy("Creating producer ");
        msg.append(i.to_string);
        msg.append(" ...");
        print_line(msg);
        !!p.make(i,num_items,data,mtx,cnd_var,barr_init,barr_finish);
        msg.copy("%N***%N");
        msg.append("*** START NEW PHASE ***%N");
        msg.append("***%N");
        print_line(msg);
        p.start;
          -- initialization barrier
        debug
          print_line("Main thread waiting on initialization barrier...");
        end;
        barr_init.wait; -- wait until all threads are initialized
        debug
          print_line("Main thread passed initialization barrier.");
        end;
        if i = num_producer_stages then
          mtx.lock;
          data.destroy;
          mtx.unlock
        end;
          -- phase finish barrier
        debug
          print_line("Main thread waiting on finalization barrier...");
        end;
        barr_finish.wait;
        debug
          print_line("Main thread passed finalization barrier.");
        end;
        i := i + 1;
      end;

      join_all -- wait until all producers & consumers terminate
    end;

end -- MAIN

