-- This  file  is   free  software,  which   comes  along  with   thread-safe
-- SmallEiffel. This software  is distributed  in the  hope that  it will  be
-- useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
-- you want, provided this  header is kept unaltered,  and a notification  of
-- the changes is  added. You  are allowed to  redistribute it  and sell  it,
-- alone or as a part of another product.
--
-- (C) Miguel Oliveira e Silva, September 2003
--

class X11_FONT

inherit
  X11_GLOBAL

creation
  make

feature

  make(disp: X11_DISPLAY) is
    require
      disp /= Void
    do
      font_var := c_x_make_font(disp.disp_var);
      loaded := false
    ensure
      not font_var.is_null
    end;

  exists(n: STRING): BOOLEAN is
    require
      n /= Void
    do
      Result := c_x_font_exists(font_var,n.to_external)
    end;

  load(n: STRING) is
    require
      n /= Void and then exists(n)
    do
      c_x_load_font(font_var,n.to_external);
      loaded := true;
      name := n
    end;

  unload is
    require
      n /= Void and then loaded
    do
      c_x_unload_font(font_var)
    end;

  loaded: BOOLEAN;

  name: STRING;

feature -- text dimensions

  compute_text_dimensions(text: STRING) is
    do
      c_x_text_extents(font_var,text.to_external,text.count,
                       $font_direction_hint,$font_ascent,$font_descent,
                       $text_lbearing,$text_rbearing,$text_width,
                       $text_ascent,$text_descent)
    ensure
      font_direction_hint = font_left_to_right or
      font_direction_hint = font_right_to_left
    end;

  font_direction_hint: INTEGER; -- font_left_to_right or font_right_to_left

  font_ascent: INTEGER; -- font's overall ascent

  font_descent: INTEGER; -- font's overall descent

  text_lbearing: INTEGER; -- origin to left edge of rectangle

  text_rbearing: INTEGER; -- origin to right edge of rectangle

  text_width: INTEGER; -- width of character string

  text_ascent: INTEGER; -- origin to top edge of rectangle

  text_descent: INTEGER; -- origin to bottom edge of rectangle

feature {X11_GRAPHIC_CONTEXT}

  font_var: POINTER;

feature {NONE}

  c_x_make_font(x11_disp: POINTER): POINTER is
    external "C"
    alias "x_make_font"
    end;

  c_x_font_exists(x11_font,n: POINTER): BOOLEAN is
    external "C"
    alias "x_font_exists"
    end;

  c_x_load_font(x11_font,n: POINTER) is
    external "C"
    alias "x_load_font"
    end;

  c_x_unload_font(x11_font: POINTER) is
    external "C"
    alias "x_unload_font"
    end;

  c_x_text_extents(x11_font,text: POINTER;text_length: INTEGER;
                   dir_hint,font_ascent,font_descent,lbearing,rbearing,
                   width,ascent,descent: POINTER) is
    external "C"
    alias "x_text_extents"
    end;

end -- X11_FONT

