-- This  file  is   free  software,  which   comes  along  with   thread-safe
-- SmallEiffel. This software  is distributed  in the  hope that  it will  be
-- useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
-- you want, provided this  header is kept unaltered,  and a notification  of
-- the changes is  added. You  are allowed to  redistribute it  and sell  it,
-- alone or as a part of another product.
--
-- (C) Miguel Oliveira e Silva, May 2003
--

class X11_GRAPHIC_CONTEXT

inherit
  X11_GLOBAL

creation
  create_new

feature

  create_new(window: X11_WINDOW) is
      -- uses default values
      -- ('create_new' name chosen due to interference with 'create' keyword)
    require
      window /= Void
    do
      gc_var := c_x_create_gc(window.display.disp_var,window.win_var);
      !!foreground_color.make_from_external(c_x_foreground(gc_var));
      !!background_color.make_from_external(c_x_background(gc_var));
    ensure
      not gc_var.is_null
    end;

  free is
    do
      gc_var := c_x_free_gc(gc_var)
    ensure
      gc_var.is_null
    end;

  foreground_color: X11_COLOR;

  set_foreground(c: X11_COLOR) is
    require
      c /= Void
    do
      foreground_color := c;
      c_x_set_foreground(gc_var,c.color_var)
    end;

  background_color: X11_COLOR;

  set_background(c: X11_COLOR) is
    require
      c /= Void
    do
      background_color := c;
      c_x_set_background(gc_var,c.color_var)
    end;

  function: INTEGER is
    do
      Result := c_x_function
    end;

  set_function(f: INTEGER) is
    require
      f >= gx_min and f <= gx_max
    do
      c_x_set_function(gc_var,f)
    end;

  line_width: INTEGER is
    external "C"
    alias "x_line_width"
    end;

  line_style: INTEGER is
    external "C"
    alias "x_line_style"
    end;

  cap_style: INTEGER is
    external "C"
    alias "x_cap_style"
    end;

  join_style: INTEGER is
    external "C"
    alias "x_join_style"
    end;

  set_line_attributes(ln_width,ln_style,ln_cap_style,ln_join_style: INTEGER) is
    require
      ln_width >= 0;
      ln_style = line_solid or ln_style = line_on_off_dash or
        ln_style = line_double_dash;
      ln_cap_style = cap_not_last or ln_cap_style = cap_butt or
        ln_cap_style = cap_round or ln_cap_style = cap_projecting;
      ln_join_style = join_miter or ln_join_style = join_round or
        ln_join_style = join_bevel;
    do
      c_x_set_line_attributes(gc_var,ln_width,ln_style,ln_cap_style,ln_join_style)
    end;

  arc_mode: INTEGER is
    external "C"
    alias "x_arc_mode"
    end;

  set_arc_mode(mode: INTEGER) is
    require
      mode = arc_chord or mode = arc_pie_slice
    do
      c_x_set_arc_mode(gc_var,mode)
    end;

  fill_style: INTEGER is
    external "C"
    alias "x_fill_style"
    end;

  set_fill_style(fs: INTEGER) is
    require
      fs = fill_solid or fs = fill_tiled or
      fs = fill_stippled or fs = fill_opaque_stippled
    do
      c_x_set_fill_style(gc_var,fs)
    end;

  fill_rule: INTEGER is
    external "C"
    alias "x_fill_rule"
    end;

  set_fill_rule(fr: INTEGER) is
    require
      fr = even_odd_rule or winding_rule
    do
      c_x_set_fill_rule(gc_var,fr)
    end;

feature -- text

  font: X11_FONT; -- default GC font if Void

  set_font(f: X11_FONT) is
    do
      font := f;
      c_x_set_font(gc_var,f.font_var)
    end;

feature {X11_GRAPHIC_CONTEXT,X11_WINDOW}

  gc_var: POINTER;

feature {NONE}

  c_x_create_gc(x11_disp,x11_win:POINTER): POINTER is
    external "C"
    alias "x_create_gc"
    end;

  c_x_free_gc(x11_gc: POINTER): POINTER is
    external "C"
    alias "x_free_gc"
    end;

  c_x_foreground(x11_gc: POINTER): POINTER is
    external "C"
    alias "x_foreground"
    end;

  c_x_set_foreground(x11_gc,x11_color: POINTER) is
    external "C"
    alias "x_set_foreground"
    end;

  c_x_background(x11_gc: POINTER): POINTER is
    external "C"
    alias "x_background"
    end;

  c_x_set_background(x11_gc,x11_color: POINTER) is
    external "C"
    alias "x_set_background"
    end;

  c_x_function: INTEGER is
    external "C"
    alias "x_function"
    end;

  c_x_set_function(x11_gc: POINTER;func: INTEGER) is
    external "C"
    alias "x_set_function"
    end;

  c_x_set_line_attributes(x11_gc:POINTER;width,ln_style,ln_cap_style,ln_join_style: INTEGER) is
    external "C"
    alias "x_set_line_attributes"
    end;

  c_x_set_arc_mode(x11_gc:POINTER;mode: INTEGER) is
    external "C"
    alias "x_set_arc_mode"
    end;

  c_x_set_fill_style(x11_gc:POINTER;fs: INTEGER) is
    external "C"
    alias "x_set_fill_style"
    end;

  c_x_set_fill_rule(x11_gc:POINTER;fr: INTEGER) is
    external "C"
    alias "x_set_fill_rule"
    end;

  c_x_set_font(x11_gc,x11_font:POINTER) is
    external "C"
    alias "x_set_font"
    end;

end -- X11_GRAPHIC_CONTEXT

