/* c_timer.c
** This  file  is   free  software,  which   comes  along  with   thread-safe
** SmallEiffel. This software  is distributed  in the  hope that  it will  be
** useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
** you want, provided this  header is kept unaltered,  and a notification  of
** the changes is  added. You  are allowed to  redistribute it  and sell  it,
** alone or as a part of another product.
**
** (C) Miguel Oliveira e Silva, August 2001
**
** thread timer
** requires POSIX.1b (and thre realtime library in linux: -lrt)
*/

#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <pthread.h>
#include "c_timer.h"

#ifdef __USE_UNIX98

static int error = 0;
static int running = 0;
static long int period = 1000L; /* default 1 second */
static int counter = 0;

static timer_t timer_id;
static pthread_mutex_t timer_mtx = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t timer_wake_up = PTHREAD_COND_INITIALIZER;

static void timer_thread(sigval_t arg)
{
  if (pthread_mutex_lock (&timer_mtx))
    error = 1;
  counter++;
  if (pthread_cond_broadcast(&timer_wake_up))
    error = 1;
  if (pthread_mutex_unlock (&timer_mtx))
    error = 1;
}

int c_counter(void)
{
  int r;

  if (pthread_mutex_lock(&timer_mtx))
    error = 1;
  r = counter;	 
  if (pthread_mutex_unlock(&timer_mtx))
    error = 1;

  return r;
}

int c_period(void)
{
  int r;

  if (pthread_mutex_lock(&timer_mtx))
    error = 1;
  r = (int)period;	 
  if (pthread_mutex_unlock(&timer_mtx))
    error = 1;

  return r;
}

void c_set_period(int msec)
{
  if (pthread_mutex_lock(&timer_mtx))
    error = 1;
  period = (long)msec;
  if (pthread_mutex_unlock(&timer_mtx))
    error = 1;
}

int c_running(void)
{
  int r;

  if (pthread_mutex_lock(&timer_mtx))
    error = 1;
  r = running;	 
  if (pthread_mutex_unlock(&timer_mtx))
    error = 1;

  return r;
}

void c_start(void)
{
  static struct itimerspec ts;
  static struct sigevent se;

  if (pthread_mutex_lock(&timer_mtx))
    error = 1;

  running = 1;
  counter = 0;

  se.sigev_notify = SIGEV_THREAD;
  se.sigev_value.sival_ptr = &timer_id;
  se.sigev_notify_function = timer_thread;
  se.sigev_notify_attributes = NULL;

  if (timer_create(CLOCK_REALTIME, &se, &timer_id) == -1)
    error = 1;

  ts.it_value.tv_sec = period / 1000L;
  ts.it_value.tv_nsec = (period - (ts.it_value.tv_sec*1000L)) * 1000000L;
  ts.it_interval.tv_sec = ts.it_value.tv_sec;
  ts.it_interval.tv_nsec = ts.it_value.tv_nsec;

  if (timer_settime(timer_id, 0, &ts, NULL) == -1)
    error = 1;

  if (pthread_mutex_unlock(&timer_mtx))
    error = 1;
}

void c_stop(void)
{
  if (pthread_mutex_lock(&timer_mtx))
    error = 1;

  running = 0;
  timer_delete(timer_id);

  if (pthread_mutex_unlock(&timer_mtx))
    error = 1;
}

int c_wait(void)
{
  int r;

  if (pthread_mutex_lock(&timer_mtx))
    error = 1;

  if (pthread_cond_wait(&timer_wake_up,&timer_mtx))
    error = 1;

  r = counter;

  if (pthread_mutex_unlock(&timer_mtx))
    error = 1;

  return r;
}

int c_error(void)
{
  return error;	  
}

#endif
