-- This  file  is   free  software,  which   comes  along  with   thread-safe
-- SmallEiffel. This software  is distributed  in the  hope that  it will  be
-- useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
-- you want, provided this  header is kept unaltered,  and a notification  of
-- the changes is  added. You  are allowed to  redistribute it  and sell  it,
-- alone or as a part of another product.
--
-- (C) Miguel Oliveira e Silva, May 2001
--
-- Virtual character terminal handler ('C' ncurses library alike)
-- (no attributes)
--

class TERMINAL_HANDLER

feature

  initialized: BOOLEAN is
    external "C"
    end;

  initialize is
    do
      if not initialized then
        init_display;
        enable_keyboard;
        show_cursor;
        unset_timeout;
        set_mouse_event_mask(empty_mouse_mask);
        set_normal_text;
        set_text_color(default_color_id);
      end
    end;

  terminate is
    do
      if initialized then
        show_cursor;
        end_display
      end
    end;

feature

  number_of_lines: INTEGER is
    external "C"
    end;
  
  number_of_columns: INTEGER is
    external "C"
    end;

feature

  cursor_visible: BOOLEAN is
    external "C"
    alias "is_cursor_visible"
    end;

  hide_cursor is
    external "C"
    end;
  
  show_cursor is
    external "C"
    end;

feature -- mouse handling

  mouse_usable: BOOLEAN is
    external "C"
    end;

  enable_mouse is
    external "C"
    end;

  disable_mouse is
    external "C"
    end;

  mouse_enabled: BOOLEAN is
    external "C"
    alias "is_mouse_enabled"
    end;

  empty_mouse_mask:            BIT 24 is 00000000000000000000B;
  button1_press_mask:          BIT 24 is 00000000000000000001B;
  button1_released_mask:       BIT 24 is 00000000000000000010B;
  button1_clicked_mask:        BIT 24 is 00000000000000000100B;
  button1_double_clicked_mask: BIT 24 is 00000000000000001000B;
  button1_triple_clicked_mask: BIT 24 is 00000000000000010000B;
  button2_press_mask:          BIT 24 is 00000000000000100000B;
  button2_released_mask:       BIT 24 is 00000000000001000000B;
  button2_clicked_mask:        BIT 24 is 00000000000010000000B;
  button2_double_clicked_mask: BIT 24 is 00000000000100000000B;
  button2_triple_clicked_mask: BIT 24 is 00000000001000000000B;
  button3_press_mask:          BIT 24 is 00000000010000000000B;
  button3_released_mask:       BIT 24 is 00000000100000000000B;
  button3_clicked_mask:        BIT 24 is 00000001000000000000B;
  button3_double_clicked_mask: BIT 24 is 00000010000000000000B;
  button3_triple_clicked_mask: BIT 24 is 00000100000000000000B;
  mouse_movement_mask:         BIT 24 is 00001000000000000000B;
  all_mouse_events_mask:       BIT 24 is 00001111111111111111B;

  valid_mouse_event_mask(mask: BIT 24): BOOLEAN is
    do
      Result := not (mask and (not all_mouse_events_mask)).to_boolean
    end;

  mouse_event_mask_usable(mask: BIT 24): BOOLEAN is
      -- is lowlevel mouse device able to handle 'mask'?
    do
      Result := is_mouse_event_mask_usable(bit_24_to_integer(mask))
    end;

  mouse_event_mask: BIT 24 is
    do
      Result := integer_to_bit_24(get_mouse_event_mask)
    end;

  set_mouse_event_mask(mask: BIT 24) is
    -- Attention: it does not require 'mouse_event_mask_usable(mask)'!
    require
      valid_mouse_event_mask(mask)
    do
      c_set_mouse_event_mask(bit_24_to_integer(mask))
    end;

feature -- keyboard handling

  enable_keyboard is
    external "C"
    end;

  disable_keyboard is
    external "C"
    end;

  keyboard_enabled: BOOLEAN is
    external "C"
    alias "is_keyboard_enabled"
    end;

feature -- especial event codes

    -- keyboard events (codes reserved from key_minimum_event_code until
    --  key_maximum_event_code)

  key_minimum_event_code: INTEGER is 0;
  key_maximum_event_code: INTEGER is 1150;

  key_none:            INTEGER is 0
  key_tab:             INTEGER is 9
  key_return:          INTEGER is 10
  key_escape:          INTEGER is 27
  key_up:              INTEGER is 1000
  key_down:            INTEGER is 1001
  key_left:            INTEGER is 1002
  key_right:           INTEGER is 1003
  key_home:            INTEGER is 1004
  key_end:             INTEGER is 1005
  key_insert:          INTEGER is 1006
  key_delete:          INTEGER is 1007
  key_page_up:         INTEGER is 1008
  key_page_down:       INTEGER is 1009
  key_back_space:      INTEGER is 1010
  key_shift_up:        INTEGER is 1011
  key_shift_down:      INTEGER is 1012
  key_shift_left:      INTEGER is 1013
  key_shift_right:     INTEGER is 1014
  key_shift_home:      INTEGER is 1015
  key_shift_end:       INTEGER is 1016
  key_shift_page_up:   INTEGER is 1017
  key_shift_page_down: INTEGER is 1018
  key_f1:              INTEGER is 1101
  key_f2:              INTEGER is 1102
  key_f3:              INTEGER is 1103
  key_f4:              INTEGER is 1104
  key_f5:              INTEGER is 1105
  key_f6:              INTEGER is 1106
  key_f7:              INTEGER is 1107
  key_f8:              INTEGER is 1108
  key_f9:              INTEGER is 1109
  key_f10:             INTEGER is 1110
  key_f11:             INTEGER is 1111
  key_f12:             INTEGER is 1112

    -- mouse events (codes reserved from mouse_minimum_event_code until
    --  mouse_maximum_event_code)

  mouse_minimum_event_code: INTEGER is 1200;
  mouse_maximum_event_code: INTEGER is 1250;

  mouse_button1_press:          INTEGER is 1200;
  mouse_button1_released:       INTEGER is 1201;
  mouse_button1_clicked:        INTEGER is 1202;
  mouse_button1_double_clicked: INTEGER is 1203;
  mouse_button1_triple_clicked: INTEGER is 1204;
  mouse_button2_press:          INTEGER is 1205;
  mouse_button2_released:       INTEGER is 1206;
  mouse_button2_clicked:        INTEGER is 1207;
  mouse_button2_double_clicked: INTEGER is 1208;
  mouse_button2_triple_clicked: INTEGER is 1209;
  mouse_button3_press:          INTEGER is 1210;
  mouse_button3_released:       INTEGER is 1211;
  mouse_button3_clicked:        INTEGER is 1212;
  mouse_button3_double_clicked: INTEGER is 1213;
  mouse_button3_triple_clicked: INTEGER is 1214;
  mouse_movement:               INTEGER is 1215;

feature

  set_timeout(msec: INTEGER) is
      -- maximum wait delay in 'fetch_event'
      -- on timeout, 'key_none' is returned.
      -- msec: negative -> infinite blocking
      --              0 -> no timeout
      --       positive -> timeout in miliseconds
    external "C"
    end
  
  unset_timeout is
      -- default behavior
      -- no timeout in 'fetch_event'
    external "C"
    end

feature

  fetch_event is
    require
      keyboard_enabled or mouse_enabled
    external "C"
    alias "get_event"
    end;

  last_event: INTEGER is
    external "C"
    end;

  keyboard_event: BOOLEAN is
      -- is last event a keyboard event?
    do
      Result := (last_event >= key_minimum_event_code) and
                (last_event <= key_maximum_event_code)
    end;

  mouse_event: BOOLEAN is
      -- is last event a mouse event?
    do
      Result := (last_event >= mouse_minimum_event_code) and
                (last_event <= mouse_maximum_event_code)
    end;

  mouse_line: INTEGER is
    -- last mouse event line position [1..number_of_lines]
    -- [1,1]: top left terminal position
    external "C"
    alias "last_mouse_y"
    end;

  mouse_column: INTEGER is
    -- last mouse event column position [1..number_of_columns]
    external "C"
    alias "last_mouse_x"
    end;

feature -- cursor

  set_cursor_position(lin,col: INTEGER) is
    require
      lin >= 1 and lin <= number_of_lines;
      col >= 1 and col <= number_of_columns;
    do
      set_cursor(lin,col)
    end;

  get_cursor_position is
    do
      get_cursor
    end;

  cursor_line: INTEGER is
    -- last 'get_cursor_position' cursor line position [1..number_of_lines]
    -- [1,1]: top left terminal position
    external "C"
    alias "last_cursor_lin"
    end;

  cursor_column: INTEGER is
    -- last 'get_cursor_position' cursor column position [1..number_of_columns]
    external "C"
    alias "last_cursor_col"
    end;

feature -- output handling

  beep is
    external "C"
    alias "do_beep"
    end;

  clear_screen is
    external "C"
    end;

  new_line is
    do
      write_char('%N')
    end;

  write_char(c: CHARACTER) is
    do
      print_char(c)
    end;

  write_char_n_times(c: CHARACTER;n: INTEGER) is
    local
      i: INTEGER
    do
      from
        i := 0
      until
        i = n
      loop
        write_char(c);
        i := i + 1
      end
    end;

  write(txt: STRING) is
    local
      i: INTEGER
    do
      from
        i := txt.lower
      until
        i = txt.upper + 1
      loop
        write_char(txt @ i);
        i := i + 1
      end
    end;

  write_bounded(txt: STRING;max_chars: INTEGER) is
    local
      i: INTEGER
    do
      from
        i := txt.lower
      until
        i - txt.lower = max_chars or i = txt.upper + 1
      loop
        write_char(txt @ i);
        i := i + 1
      end
    end;

  writeln(txt: STRING) is
    do
      write(txt);
      new_line
    end;

  write_char_at(lin,col: INTEGER;c: CHARACTER) is
    require
      lin >= 1 and lin <= number_of_lines;
      col >= 1 and col <= number_of_columns;
    do
      set_cursor_position(lin,col);
      write_char(c)
    end;

  write_at(lin,col: INTEGER;txt: STRING) is
    require
      lin >= 1 and lin <= number_of_lines;
      col >= 1 and col <= number_of_columns;
    do
      set_cursor_position(lin,col);
      write(txt)
    end;

  write_bounded_at(lin,col: INTEGER;txt: STRING;max_chars: INTEGER) is
    require
      lin >= 1 and lin <= number_of_lines;
      col >= 1 and col <= number_of_columns;
    do
      set_cursor_position(lin,col);
      write_bounded(txt,max_chars)
    end;

  writeln_at(lin,col: INTEGER;txt: STRING) is
    require
      lin >= 1 and lin <= number_of_lines;
      col >= 1 and col <= number_of_columns;
    do
      write_at(lin,col,txt);
      new_line
    end;

feature -- write properties

  text_attributes: BIT 24 is
    do
      Result := integer_to_bit_24(text_attributes_ref.item)
    end;

  set_text_attributes(a: BIT 24) is
    do
      store_text_attributes(bit_24_to_integer(a));
      if a = normal_text_id then
        c_set_normal_text
      else
        if (a and highlight_id).to_boolean then
          c_set_highlight
        else
          c_reset_highlight
        end;
        if (a and underline_id).to_boolean then
          c_set_underline
        else
          c_reset_underline
        end;
        if (a and reverse_id).to_boolean then
          c_set_reverse
        else
          c_reset_reverse
        end;
        if (a and blinking_id).to_boolean then
          c_set_blinking
        else
          c_reset_blinking
        end;
        if (a and dim_id).to_boolean then
          c_set_dim
        else
          c_reset_dim
        end;
        if (a and bold_id).to_boolean then
          c_set_bold
        else
          c_reset_bold
        end;
      end
    end;

  normal_text_id: BIT 24 is 00000000B;
  highlight_id:   BIT 24 is 00000001B;
  underline_id:   BIT 24 is 00000010B;
  reverse_id:     BIT 24 is 00000100B;
  blinking_id:    BIT 24 is 00001000B;
  dim_id:         BIT 24 is 00010000B;
  bold_id:        BIT 24 is 00100000B;

  set_normal_text is
      -- Normal display (no highlight)
    do
      store_text_attributes(bit_24_to_integer(normal_text_id));
      c_set_normal_text
    end;

  set_highlight is
      -- Best highlighting mode of the terminal
    do
      store_text_attributes(bit_24_to_integer(text_attributes or highlight_id));
      c_set_highlight
    end;

  reset_highlight is
    do
      store_text_attributes(bit_24_to_integer(text_attributes and (not highlight_id)));
      c_reset_highlight
    end;

  set_underline is
    do
      store_text_attributes(bit_24_to_integer(text_attributes or underline_id));
      c_set_underline
    end;

  reset_underline is
    do
      store_text_attributes(bit_24_to_integer(text_attributes and (not underline_id)));
      c_reset_underline
    end;

  set_reverse is
    do
      store_text_attributes(bit_24_to_integer(text_attributes or reverse_id));
      c_set_reverse
    end;

  reset_reverse is
    do
      store_text_attributes(bit_24_to_integer(text_attributes and (not reverse_id)));
      c_reset_reverse
    end;

  set_blinking is
    do
      store_text_attributes(bit_24_to_integer(text_attributes or blinking_id));
      c_set_blinking
    end;

  reset_blinking is
    do
      store_text_attributes(bit_24_to_integer(text_attributes and (not blinking_id)));
      c_reset_blinking
    end;

  set_dim is
      -- Half bright
    do
      store_text_attributes(bit_24_to_integer(text_attributes or dim_id));
      c_set_dim
    end;

  reset_dim is
    do
      store_text_attributes(bit_24_to_integer(text_attributes and (not dim_id)));
      c_reset_dim
    end;

  set_bold is
      -- Extra bright
    do
      store_text_attributes(bit_24_to_integer(text_attributes or bold_id));
      c_set_bold
    end;

  reset_bold is
    do
      store_text_attributes(bit_24_to_integer(text_attributes and (not bold_id)));
      c_reset_bold
    end;

feature -- color

  text_color: INTEGER is
    do
      Result := text_color_ref.item
    end;

  set_text_color(c: INTEGER) is
    do
      store_text_color(c);
      inspect c
        when default_color_id then
          c_set_black_color
        when black_color_id then
          c_set_black_color
        when red_color_id then
          c_set_red_color
        when green_color_id then
          c_set_green_color
        when yellow_color_id then
          c_set_yellow_color
        when blue_color_id then
          c_set_blue_color
        when cyan_color_id then
          c_set_cyan_color
        when magenta_color_id then
          c_set_magenta_color
        when white_color_id then
          c_set_white_color
      end
    end;

  default_color_id: INTEGER is unique;
  black_color_id: INTEGER is unique;
  red_color_id: INTEGER is unique;
  green_color_id: INTEGER is unique;
  yellow_color_id: INTEGER is unique;
  blue_color_id: INTEGER is unique;
  cyan_color_id: INTEGER is unique;
  magenta_color_id: INTEGER is unique;
  white_color_id: INTEGER is unique;

  set_black_color is
    do
      store_text_color(black_color_id);
      c_set_black_color
    end;

  set_red_color is
    do
      store_text_color(red_color_id);
      c_set_red_color
    end;

  set_green_color is
    do
      store_text_color(green_color_id);
      c_set_green_color
    end;

  set_yellow_color is
    do
      store_text_color(yellow_color_id);
      c_set_yellow_color
    end;

  set_blue_color is
    do
      store_text_color(blue_color_id);
      c_set_blue_color
    end;

  set_cyan_color is
    do
      store_text_color(cyan_color_id);
      c_set_cyan_color
    end;

  set_magenta_color is
    do
      store_text_color(magenta_color_id);
      c_set_magenta_color
    end;

  set_white_color is
    do
      store_text_color(white_color_id);
      c_set_white_color
    end;

--feature -- window save/restore
--
--  fetch_area(xi,yi,xf,yf: INTEGER;area: CHARACTER_WINDOW) is
--    do
--    end;
--
--  restore_area(xi,yi: INTEGER;area: CHARACTER_WINDOW) is
--    do
--    end;

feature -- advanced features

  edit_line_text(lin,col,width: INTEGER; text: STRING;
                 fixed: INTEGER; cursor_at_end: BOOLEAN): BOOLEAN is
      -- true result if text changed (this is bad because it does not obey CQS).
      -- (lin,col): leftmost position of edit window
      -- width: width edit window
      -- text: string to edit
      -- fixed: number of initial 'text' characters to remain unchangeable
    require
      lin >= 1 and lin <= number_of_lines;
      col >= 1 and col <= number_of_columns;
      width > 0 and width <= number_of_columns - col + 1;
      text /= Void and then (fixed >= 0 and fixed <= text.count)
    local
      start_chr,cursor_chr,k: INTEGER;
      exit_loop,overwrite_mode,is_cur_visible: BOOLEAN;
      cpy: STRING;
    do
      !!cpy.make(text.capacity);
      cpy.copy(text);

      is_cur_visible := cursor_visible;

      if not is_cur_visible then
        show_cursor
      end;

      if cursor_at_end then
        start_chr := cpy.upper + 1 - width;
        if start_chr < cpy.lower then
          start_chr := cpy.lower
        end;
        cursor_chr := cpy.upper + 1
      else
        start_chr := cpy.lower;
        cursor_chr := cpy.lower
      end;
      print_visible_text(lin,col,width,cpy,start_chr,cursor_chr);
      from
        overwrite_mode := false;
        exit_loop := false
      until
        exit_loop
      loop
        fetch_event;
        inspect last_event
          when key_return then
            if not text.is_equal(cpy) then
              text.copy(cpy);
              Result := true
            end;
            exit_loop := true
          when key_escape then
            exit_loop := true
          when key_insert then
            overwrite_mode := not overwrite_mode
          when key_left then
            if cursor_chr > cpy.lower then
              cursor_chr := cursor_chr - 1;
              if start_chr > cursor_chr then
                start_chr := cursor_chr;
                print_visible_text(lin,col,width,cpy,start_chr,cursor_chr)
              else
                set_cursor_position(lin,col + (cursor_chr - start_chr))
              end
            end
          when key_right then
            if cursor_chr <= cpy.upper then
              cursor_chr := cursor_chr + 1;
              if start_chr < cursor_chr - width then
                start_chr := cursor_chr - width;
                print_visible_text(lin,col,width,cpy,start_chr,cursor_chr)
              else
                set_cursor_position(lin,col + (cursor_chr - start_chr))
              end
            end
          when key_home then
            start_chr := cpy.lower;
            cursor_chr := cpy.lower;
            print_visible_text(lin,col,width,cpy,start_chr,cursor_chr)
          when key_end then
            start_chr := cpy.upper + 1 - width;
            if start_chr < cpy.lower then
              start_chr := cpy.lower
            end;
            cursor_chr := cpy.upper + 1;
            print_visible_text(lin,col,width,cpy,start_chr,cursor_chr)
          when key_delete then
            if cursor_chr <= cpy.upper and cursor_chr >= fixed + cpy.lower then
              cpy.remove(cursor_chr);
              print_visible_text(lin,col,width,cpy,start_chr,cursor_chr)
            else
              beep
            end
          when key_back_space then
            if cursor_chr > cpy.lower and cursor_chr > fixed + cpy.lower then
              cpy.remove(cursor_chr - 1);
              cursor_chr := cursor_chr - 1;
              if start_chr > cursor_chr then
                start_chr := cursor_chr
              end;
              print_visible_text(lin,col,width,cpy,start_chr,cursor_chr)
            else
              beep
            end
          when mouse_button1_clicked then
            beep;
            if mouse_line = lin and
               mouse_column >= col and mouse_column < col + width then
              cursor_chr := start_chr + mouse_column - col;
              set_cursor_position(lin,col + (cursor_chr - start_chr))
            end
          when key_none then
            -- does nothing
          else
--            set_cursor_position(25,1);
--            write("last event: ");
--            write(last_event.to_string);
            if cpy.count < cpy.capacity and cursor_chr >= fixed + cpy.lower then
              if (last_event >= 32 and last_event <= 126) or
                 (last_event >= 160 and last_event <= 255) then
                if overwrite_mode and cursor_chr <= cpy.upper then
                  cpy.put(last_event.to_character,cursor_chr)
                else
                  cpy.insert(last_event.to_character,cursor_chr - 1)
                end;
                cursor_chr := cursor_chr + 1;
                if start_chr < cursor_chr - width then
                  start_chr := cursor_chr - width;
                end;
                print_visible_text(lin,col,width,cpy,start_chr,cursor_chr)
              end
            else
              beep
            end
        end
      end;

      if not is_cur_visible then
        hide_cursor
      end;
    end;

feature {NONE}

  print_visible_text(lin,col,width: INTEGER;text: STRING;
                     start_chr,cursor_chr: INTEGER) is
    local
      i: INTEGER
    do
      hide_cursor;
      set_cursor_position(lin,col);
      from
        i := start_chr
      until
        i - start_chr = width or i - text.lower = text.count
      loop
        write_char(text @ i);
        i := i + 1
      end;
      write_char_n_times(' ',width - (i - start_chr));
      show_cursor;
      set_cursor_position(lin,col + (cursor_chr - start_chr));
    end;

  init_display is
    external "C"
    end;

  end_display is
    external "C"
    end;

  get_mouse_event_mask: INTEGER is
    external "C"
    end;

  c_set_mouse_event_mask(m: INTEGER) is
    external "C"
    alias "set_mouse_event_mask"
    end;

  is_mouse_event_mask_usable(m: INTEGER): BOOLEAN is
    external "C"
    end;

  last_mouse_x: INTEGER is
    external "C"
    end;

  last_mouse_y: INTEGER is
    external "C"
    end;

  set_cursor(lin,col: INTEGER) is
    external "C"
    end;

  get_cursor is
    external "C"
    end;

  last_cursor_lin:INTEGER is
    external "C"
    end;

  last_cursor_col:INTEGER is
    external "C"
    end;

  print_char(c: CHARACTER) is
    external "C"
    end;

  store_text_attributes(a: INTEGER) is
    do
      text_attributes_ref.set_item(a)
    end;

  store_text_color(c: INTEGER) is
    do
      text_color_ref.set_item(c)
    end;

  c_set_normal_text is
    external "C"
    alias "set_normal_text"
    end;

  c_set_highlight is
    external "C"
    alias "set_highlight"
    end;

  c_reset_highlight is
    external "C"
    alias "reset_highlight"
    end;

  c_set_underline is
    external "C"
    alias "set_underline"
    end;

  c_reset_underline is
    external "C"
    alias "reset_underline"
    end;

  c_set_reverse is
    external "C"
    alias "set_reverse"
    end;

  c_reset_reverse is
    external "C"
    alias "reset_reverse"
    end;

  c_set_blinking is
    external "C"
    alias "set_blinking"
    end;

  c_reset_blinking is
    external "C"
    alias "reset_blinking"
    end;

  c_set_dim is
    external "C"
    alias "set_dim"
    end;

  c_reset_dim is
    external "C"
    alias "reset_dim"
    end;

  c_set_bold is
    external "C"
    alias "set_bold"
    end;

  c_reset_bold is
    external "C"
    alias "reset_bold"
    end;

  c_set_black_color is
    external "C"
    alias "set_black_color"
    end;

  c_set_red_color is
    external "C"
    alias "set_red_color"
    end;

  c_set_green_color is
    external "C"
    alias "set_green_color"
    end;

  c_set_yellow_color is
    external "C"
    alias "set_yellow_color"
    end;

  c_set_blue_color is
    external "C"
    alias "set_blue_color"
    end;

  c_set_cyan_color is
    external "C"
    alias "set_cyan_color"
    end;

  c_set_magenta_color is
    external "C"
    alias "set_magenta_color"
    end;

  c_set_white_color is
    external "C"
    alias "set_white_color"
    end;

  text_attributes_ref: INTEGER_REF is
    once
      !!Result
    end;

  text_color_ref: INTEGER_REF is
    once
      !!Result
    end;

  integer_to_bit_24(i: INTEGER): BIT 24 is
    local
      c,j: INTEGER
    do
      from
        j := i;
        c := 24
      until
        c = 0
      loop
        if j \\ 2 = 1 then
          Result.put_1(c)
        else
          Result.put_0(c)
        end;
        j := j // 2;
        c := c - 1
      end
    end;

  bit_24_to_integer(b: BIT 24): INTEGER is
    local
      c: INTEGER
    do
      Result := 0;
      from
        c := 1
      until
        c > 24
      loop
        Result := Result * 2;
        if b.item(c) then
          Result := Result + 1
        end;
        c := c + 1
      end
    end;

end -- TERMINAL_HANDLER

