-- This  file  is   free  software,  which   comes  along  with   thread-safe
-- SmallEiffel. This software  is distributed  in the  hope that  it will  be
-- useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
-- you want, provided this  header is kept unaltered,  and a notification  of
-- the changes is  added. You  are allowed to  redistribute it  and sell  it,
-- alone or as a part of another product.
--
-- (C) Miguel Oliveira e Silva, August 2001
--
-- Changed to new global once in August 2003
--
-- Thread-safe global timer class
--

class TIMER

creation
  make

feature

  make is
      -- the first calling thread becomes the master thread (always)
    local
      tid: THREAD_ID
    do
      tid := master_thread_id;
    end;

feature

  period: INTEGER is
      -- [miliseconds]
      -- default: 1000 miliseconds (1 sec)
    external "C"
    alias "c_period"
    end;

  set_period(msec: INTEGER) is
    require
      not running;
      is_master_thread
    external "C"
    alias "c_set_period"
    end;

  running: BOOLEAN is
    external "C"
    alias "c_running"
    end;

  start is
    require
      not running;
      is_master_thread
    external "C"
    alias "c_start"
    end;

  stop is
    require
      running;
      is_master_thread
    external "C"
    alias "c_stop"
    end;

  wait is
      -- wait for next period timeout
    require
      running
    local
      c: INTEGER
    do
      c := c_wait;
      if c > last_counter_ref.item + 1 then
        num_lost_periods_ref.set_item(num_lost_periods_ref.item +
                                      c - last_counter_ref.item - 1)
      end;
      last_counter_ref.set_item(c);
    end;

  counter: INTEGER is
      -- per-thread
    do
      Result := last_counter_ref.item
    end;

  is_master_thread: BOOLEAN is
    do
      Result := master_thread_id.same_as(curr_thread_id)
    end;

feature -- periods passed without explicit waits

  number_of_lost_periods: INTEGER is
    do
      Result := num_lost_periods_ref.item
    end;

  reset_lost_periods is
    do
      num_lost_periods_ref.set_item(0);
      last_counter_ref.set_item(c_counter);
    end;

  error: BOOLEAN is
    external "C"
    alias "c_error"
    end;

feature {NONE}

  curr_thread_id: THREAD_ID is
    once
      !!Result.make
    end;

  last_counter_ref: INTEGER_REF is
    once
      !!Result
    end;

  num_lost_periods_ref: INTEGER_REF is
    once
      !!Result
    end;

  master_thread_id: THREAD_ID is
    once {"process"}
      !!Result.make
    end;

  c_counter: INTEGER is
    external "C"
    end;

  c_wait: INTEGER is
    external "C"
    end;

invariant

  not error

end -- TIMER

