--
-- (C) Miguel Oliveira e Silva (mos@det.ua.pt), August 2001
--
-- Normalized gravitation:
--   - 1 meter = 0.5 line = 1 column
--   - the gravitation constant (6.67E-11) is normalized to 1.0
--

class MAIN

inherit
  THREAD_CONTROL;
  THREAD_TERMINAL_HANDLER

creation
  make

feature

  maximum_particles: INTEGER is 50;

  period: INTEGER is 25; -- refresh ratio in ms

  time_increment: DOUBLE is 0.005; -- simulation time increment in seconds

  num_particles: INTEGER;

  particle_list: PARTICLE_LIST;

  make is
    local
      i,j,update_freq: INTEGER;
      finish: BOOLEAN;
      p0,v0: VECTOR;
      inc,d,m: DOUBLE;
      part: PARTICLE;
      timer: TIMER;
      barr: THREAD_BARRIER;
      clk: CLOCK;
    do
      if argument_count /= 1 then
        print_string("Usage: a.out <num-planets>%N")
      elseif not argument(1).is_integer or else argument(1).to_integer <= 0 then
        print_string("ERROR: expecting an positive integer argument!%N")
      elseif argument(1).to_integer > maximum_particles then
        print_string("ERROR: maximum number of particles exceeded (");
        print_integer(maximum_particles);
        print_string(")!%N")
      else
        num_particles := argument(1).to_integer - 1;
        initialize; -- character terminal
        hide_cursor;
        enable_keyboard;
        clear_screen;

        !!particle_list.make(maximum_particles);

        !!barr.make;
        barr.sign_on;

        !!timer.make;
        timer.set_period(period);
        timer.start;

        !!clk.make(timer);
        clk.start;

        !!p0;
        !!v0;

          -- star

           -- mass
        m := 1.0E5;
           -- position
        p0.set_x(number_of_columns // 2);
        p0.set_y(number_of_lines);
           -- velocity
        v0.set_zero;
        !!part.make(m,p0,v0,time_increment,timer,barr,particle_list);
        particle_list.add(part);

          -- planets

        inc := number_of_lines.to_double / num_particles / 1.5;
        from
          d := 0.0;
          i := 1
        until
          i > num_particles
        loop
          d := d + inc;
          p0.set_x((number_of_columns // 2).to_double + d);
          p0.set_y(number_of_lines.to_double);
          v0.set_x(0.0);
          v0.set_y((m / d).sqrt);
          !!part.make(1.0,p0,v0,time_increment,timer,barr,particle_list);
          particle_list.add(part);
          i := i + 1
        end;

        particle_list.read_lock;
        from
          i := 1
        until
          i > particle_list.count
        loop
          part := particle_list @ i;
          if part /= Void then
            part.start
          end;
          i := i + 1
        end;
        particle_list.read_unlock;

        update_freq := 1000 // timer.period;
        from
          j := 0
        until
          finish
        loop
          timer.wait;
          if j = update_freq then
            write_at(number_of_lines,1,"(exit: Esc)");
            j := 0
          else
            j := j + 1
          end;
          fetch_event;
          finish := keyboard_event and then last_event = key_escape;

          if finish then -- signal termination to all particles
            clk.terminate_clock;
            particle_list.read_lock;
            from
              i := 1
            until
              i > particle_list.count
            loop
              part := particle_list @ i;
              if part /= Void and then not part.finish then
                part.terminate_particle
              end;
              i := i + 1
            end;
            particle_list.read_unlock;
            barr.terminate;
          else
            barr.wait;
          end;
        end;

        join_all;
        clear_screen;
        show_cursor;
        terminate -- character terminal
      end
    end;

end -- MAIN

