--
-- (C) Miguel Oliveira e Silva (mos@det.ua.pt), August 2001
--

class PARTICLE_LIST

creation
  make

feature

  make(max_size: INTEGER) is
    do
      !!rwl.make_with_write_priority;
      !!list.make(1,max_size);
      size := 0;
    end;

  add(p: PARTICLE) is
    require
      not full;
      not contains(p)
    do
      write_lock;
      size := size + 1;
      list.put(p,size);
      write_unlock;
    end;

  remove(p: PARTICLE) is
    require
      not empty;
      contains(p)
    local
      i: INTEGER
    do
      write_lock;
      from
        i := 1
      until
        (list @ i) = p
      loop
        i := i + 1
      end;
      from
      until
        i = size
      loop
        list.put(list @ (i + 1),i);
        i := i + 1
      end;
      size := size - 1;
      write_unlock;
    end;

  remove_if_present(p: PARTICLE) is
    local
      i: INTEGER
    do
      write_lock;
      from
        i := 1
      until
        i > size or else (list @ i) = p
      loop
        i := i + 1
      end;
      if i <= size then
        from
        until
          i = size
        loop
          list.put(list @ (i + 1),i);
          i := i + 1
        end;
        size := size - 1;
      end;
      write_unlock;
    end;

  empty: BOOLEAN is
    do
      read_lock;
      Result := size = 0;
      read_unlock;
    end;

  full: BOOLEAN is
    do
      read_lock;
      Result := size = list.upper;
      read_unlock;
    end;

  contains(p: PARTICLE): BOOLEAN is
    local
      i: INTEGER
    do
      read_lock;
      from
        i := 1
      until
        i > size or else (list @ i) = p
      loop
        i := i + 1
      end;
      Result := i <= size;
      read_unlock;
    end;

feature -- unsafe features that must be explicitly protected with read locks.

  count: INTEGER is
    do
      Result := size
    end;

  item, infix "@" (i: INTEGER): PARTICLE is
    require
      i >= 1 and i <= count
    do
      Result := (list @ i)
    end;

feature {NONE}

  list: ARRAY[PARTICLE];

  size: INTEGER;

feature

  read_lock is
    do
      rwl.read_lock
    end;

  read_unlock is
    do
      rwl.read_unlock
    end;

  write_lock is
    do
      rwl.write_lock
    end;

  write_unlock is
    do
      rwl.write_unlock
    end;

feature {NONE}

  rwl: READ_WRITE_LOCK;

end -- PARTICLE_LIST

