--
-- (C) Miguel Oliveira e Silva (mos@det.ua.pt), August-October 2000
--

class PRODUCER

inherit
  THREAD

creation
  make

feature {PRODUCER}

  main is
    local
      j,r: INTEGER;
      str: STRING;
      i: ITEM;
      msg: STRING
    do
      !!msg.make(80);
      msg.copy("Producer ");
      msg.append(id.to_string);
      msg.append(" started.");
      print_line(msg);
      data_mutex.lock;
      data.activate_producer;
      data_mutex.unlock;
      from
        j := 1
      until
        j > num_items
      loop
        from
          r := random_integer(5);
        until
          r = 0 or j > num_items
        loop
          !!str.make(16);
          str.copy("produced by ");
          str.append(id.to_string);
          !!i;
          i.set(j,str);
          data_mutex.lock; -- serialize shared data access
          data.add(i);
          cond_var.signal; -- notify new entry to [any] consumer
          data_mutex.unlock;
          j := j + 1;
          r := r - 1
        end;
        us_sleep(1000*random_integer(1000));
      end;
      data_mutex.lock;
      data.deactivate_producer; -- notify producer termination in shared
                                -- data structure
      if id = 1 then -- one producer needs to guarantee that all
                     -- SHARED_DATA entries are consumed
        from
        until
          data.empty and data.finish
        loop
          cond_var.signal;
          data_mutex.unlock;
          ns_sleep(1); -- a little unlocked time to allow a lock from
                       -- [any] consumer
          data_mutex.lock
        end;
        cond_var.broadcast; -- making sure no consumer will be waiting forever 
      end;

      data_mutex.unlock;
      msg.copy("Producer ");
      msg.append(id.to_string);
      msg.append(" finish.");
      print_line(msg);
    end;

feature

  make(tid: INTEGER;nitems: INTEGER;dt: SHARED_DATA;mtx: MUTEX;cvar: CONDITION_VARIABLE) is
    do
      id := tid;
      num_items := nitems;
      data := dt;
      data_mutex := mtx;
      cond_var := cvar
    end;

  id: INTEGER;

  num_items: INTEGER;

  data: SHARED_DATA;

  data_mutex: MUTEX;

  cond_var: CONDITION_VARIABLE;

end -- PRODUCER

