-- This  file  is   free  software,  which   comes  along  with   thread-safe
-- SmallEiffel. This software  is distributed  in the  hope that  it will  be
-- useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
-- you want, provided this  header is kept unaltered,  and a notification  of
-- the changes is  added. You  are allowed to  redistribute it  and sell  it,
-- alone or as a part of another product.
--
-- (C) Miguel Oliveira e Silva (mos@det.ua.pt), August 2000
--

class CONDITION_VARIABLE

creation
  make

feature

  initialized: BOOLEAN;

  make is
    do
      cond_var_state := c_initialize_cond_var_state;
      initialized := true
    end;

  destroy is
    -- destroys condition variable
    do
      c_finalize_cond_var_state(cond_var_state);
      cond_var_state := Void;
      initialized := false
    end;

  wait(m: MUTEX) is
      -- m must be locked
    do
      c_cond_var_wait(cond_var_state,m.mutex_state);
    end;

  timedwait(m: MUTEX;timeout: INTEGER): BOOLEAN is
    -- Returns false on timeout, and true if signaled
    --  timeout is the absolute time in seconds (relative
    --  to 00:00:00 GMT, January 1, 1970)
    -- Absolute time is used, instead of elapsed time, because
    --  of spurious wakenings (always possible with cond. variables).
    do
      Result := c_cond_var_timedwait(cond_var_state,m.mutex_state,timeout);
    end;

  signal is
    do
      c_cond_var_signal(cond_var_state);
    end;

  broadcast is
    do
      c_cond_var_broadcast(cond_var_state);
    end;

feature {NONE}

  cond_var_state: POINTER;
    -- internal cond_var state

  frozen c_initialize_cond_var_state: POINTER is
    external "C"
    alias "initialize_cond_var_state"
    end;

  frozen c_finalize_cond_var_state(cvt: POINTER) is
    external "C"
    alias "finalize_cond_var_state"
    end;

  frozen cond_var_error(cvt: POINTER): BOOLEAN is
    external "C"
    end;

  frozen c_cond_var_wait(cvt: POINTER;mt: POINTER) is
    external "C"
    alias "cond_var_wait"
    end;

  frozen c_cond_var_timedwait(cvt: POINTER;mt: POINTER;timeout: INTEGER): BOOLEAN is
    external "C"
    alias "cond_var_timedwait"
    end;

  frozen c_cond_var_signal(cvt: POINTER) is
    external "C"
    alias "cond_var_signal"
    end;

  frozen c_cond_var_broadcast(cvt: POINTER) is
    external "C"
    alias "cond_var_broadcast"
    end;

invariant

  not cond_var_error(cond_var_state)

end -- CONDITION_VARIABLE

