-- This  file  is   free  software,  which   comes  along  with   thread-safe
-- SmallEiffel. This software  is distributed  in the  hope that  it will  be
-- useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
-- you want, provided this  header is kept unaltered,  and a notification  of
-- the changes is  added. You  are allowed to  redistribute it  and sell  it,
-- alone or as a part of another product.
--
-- (C) Miguel Oliveira e Silva (mos@det.ua.pt), August 2000
--

class MUTEX

creation
  make

feature

  initialized: BOOLEAN;

  make is
    do
      mutex_state := c_initialize_mutex_state;
      initialized := true
    end;

  destroy is
    -- destroys mutex
    do
      c_finalize_mutex_state(mutex_state);
      mutex_state := Void;
      initialized := false
    end;

  lock is
    do
      c_mutex_lock(mutex_state);
    end;

  try_lock: BOOLEAN is
    -- on lock success returns true (false otherwise)
    do
      Result := c_mutex_trylock(mutex_state);
    end;

  unlock is
    do
      c_mutex_unlock(mutex_state);
    end;

feature {CONDITION_VARIABLE}

  mutex_state: POINTER;
    -- internal mutex state

feature {NONE}

  frozen c_initialize_mutex_state: POINTER is
    external "C"
    alias "initialize_mutex_state"
    end;

  frozen c_finalize_mutex_state(mt: POINTER) is
    external "C"
    alias "finalize_mutex_state"
    end;

  frozen mutex_error(mt: POINTER): BOOLEAN is
    external "C"
    end;

  frozen c_mutex_lock(mt: POINTER) is
    external "C"
    alias "mutex_lock"
    end;

  frozen c_mutex_trylock(mt: POINTER): BOOLEAN is
    external "C"
    alias "mutex_trylock"
    end;

  frozen c_mutex_unlock(mt: POINTER) is
    external "C"
    alias "mutex_unlock"
    end;

invariant

  not mutex_error(mutex_state)

end -- MUTEX

