-- This  file  is   free  software,  which   comes  along  with   thread-safe
-- SmallEiffel. This software  is distributed  in the  hope that  it will  be
-- useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
-- you want, provided this  header is kept unaltered,  and a notification  of
-- the changes is  added. You  are allowed to  redistribute it  and sell  it,
-- alone or as a part of another product.
--
-- (C) Miguel Oliveira e Silva (mos@det.ua.pt), August 2001
--
-- eiffel interface to new POSIX rwlock pthread library
--

class READ_WRITE_LOCK

creation
  make,make_with_write_priority,make_with_read_priority

feature

  make,make_with_write_priority is
    do
      rwl_state := c_initialize_rwl_state_with_write_priority;
    end;

  make_with_read_priority is
    do
      rwl_state := c_initialize_rwl_state_with_read_priority;
    end;

  destroy is
    do
      c_finalize_rwl_state(rwl_state);
      rwl_state := Void;
    end;

  read_lock is
    do
      c_rwl_readlock(rwl_state);
    end;

  read_try_lock: BOOLEAN is
    -- on lock success returns true (false otherwise)
    do
      Result := c_rwl_readtrylock(rwl_state);
    end;

  read_unlock is
    do
      c_rwl_readunlock(rwl_state);
    end;

  write_lock is
    do
      c_rwl_writelock(rwl_state);
    end;

  write_try_lock: BOOLEAN is
    -- on lock success returns true (false otherwise)
    do
      Result := c_rwl_writetrylock(rwl_state);
    end;

  write_unlock is
    do
      c_rwl_writeunlock(rwl_state);
    end;

  write_lock_priority: BOOLEAN is
    do
      Result := c_write_lock_priority;
    end;

  read_lock_priority: BOOLEAN is
    do
      Result := not c_write_lock_priority;
    end;

feature {READ_WRITE_LOCK}

  rwl_state: POINTER;
    -- internal rwl state

feature {NONE}

  frozen c_initialize_rwl_state_with_write_priority: POINTER is
    external "C"
    alias "initialize_rwl_state_with_write_priority"
    end;

  frozen c_initialize_rwl_state_with_read_priority: POINTER is
    external "C"
    alias "initialize_rwl_state_with_read_priority"
    end;

  frozen c_finalize_rwl_state(rwl: POINTER) is
    external "C"
    alias "finalize_rwl_state"
    end;

  frozen rwl_error(rwl: POINTER): BOOLEAN is
    external "C"
    end;

  frozen c_rwl_readlock(rwl_s: POINTER) is
    external "C"
    end;

  frozen c_rwl_readtrylock(rwl_s: POINTER): BOOLEAN is
    external "C"
    end;

  frozen c_rwl_readunlock(rwl_s: POINTER) is
    external "C"
    end;

  frozen c_rwl_writelock(rwl_s: POINTER) is
    external "C"
    end;

  frozen c_rwl_writetrylock(rwl_s: POINTER): BOOLEAN is
    external "C"
    end;

  frozen c_rwl_writeunlock(rwl_s: POINTER) is
    external "C"
    end;

  frozen c_write_lock_priority(rwl: POINTER): BOOLEAN is
    external "C"
    end;

invariant

  not rwl_error(rwl_state)

end -- READ_WRITE_LOCK

