-- This  file  is   free  software,  which   comes  along  with   thread-safe
-- SmallEiffel. This software  is distributed  in the  hope that  it will  be
-- useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
-- you want, provided this  header is kept unaltered,  and a notification  of
-- the changes is  added. You  are allowed to  redistribute it  and sell  it,
-- alone or as a part of another product.
--
-- (C) Miguel Oliveira e Silva (mos@det.ua.pt), August 2001
--
-- Eiffel generic implementation of a read-write lock
--
-- (used when the POSIX thread library doen't have rwlocks)
--

class READ_WRITE_LOCK

creation
  make,make_with_write_priority,make_with_read_priority

feature

  make,make_with_write_priority is
    do
      initialize;
      w_lock_priority := true;
    end;

  make_with_read_priority is
    do
      initialize;
      w_lock_priority := false;
    end;

  destroy is
    do
    end;

  read_lock is
    do
      mtx.lock;
      if w_active or (w_lock_priority and w_wait > 0) then
        r_wait := r_wait + 1;
        from
        until
          not (w_active or (w_lock_priority and w_wait > 0))
        loop
          read.wait(mtx);
        end;
        r_wait := r_wait - 1
      end;
      r_active := r_active + 1;
      mtx.unlock
    end;

  read_try_lock: BOOLEAN is
    -- on lock success returns true (false otherwise)
    do
      mtx.lock;
      Result := not (w_active or (w_lock_priority and w_wait > 0));
      if Result then
        r_active := r_active + 1
      end;
      mtx.unlock
    end;

  read_unlock is
    do
      mtx.lock;
      r_active := r_active - 1
      if r_active = 0 and w_wait > 0 then
        write.signal
      end;
      mtx.unlock
    end;

  write_lock is
    do
      mtx.lock;
      if w_active or r_active > 0 then
          w_wait := w_wait + 1;
          from
          until
            not (w_active or r_active > 0)
          loop
            write.wait(mtx)
          end;
          w_wait := w_wait - 1
      end;
      w_active := true;
      mtx.unlock
    end;

  write_try_lock: BOOLEAN is
    -- on lock success returns true (false otherwise)
    do
      mtx.lock;
      Result := not (w_active or r_active > 0);
      w_active := Result;
      mtx.unlock
    end;

  write_unlock is
    do
      mtx.lock;
      w_active := false;
      if w_lock_priority then
        if w_wait > 0 then
          write.signal;
        elseif r_wait > 0 then
          read.broadcast;
        end
      else
        if r_wait > 0 then
          read.broadcast;
        elseif w_wait > 0 then
          write.signal;
        end
      end;
      mtx.unlock
    end;

  write_lock_priority: BOOLEAN is
    do
      mtx.lock;
      Result := w_lock_priority;
      mtx.unlock
    end;

  read_lock_priority: BOOLEAN is
    do
      mtx.lock;
      Result := not w_lock_priority;
      mtx.unlock
    end;

feature {NONE}

  initialize is
    do
      !!mtx.make;
      !!read.make;
      !!write.make;
      r_active := 0;
      w_active := false;
      r_wait := 0;
      w_wait := 0;
    end;

  mtx: MUTEX;

  read, write: CONDITION_VARIABLE;

  r_active: INTEGER; -- readers active

  w_active: BOOLEAN; -- writer active

  r_wait: INTEGER; -- readers waiting

  w_wait: INTEGER; -- writers waiting

  w_lock_priority: BOOLEAN -- write or read priority;

end -- READ_WRITE_LOCK

