-- This  file  is   free  software,  which   comes  along  with   thread-safe
-- SmallEiffel. This software  is distributed  in the  hope that  it will  be
-- useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
-- you want, provided this  header is kept unaltered,  and a notification  of
-- the changes is  added. You  are allowed to  redistribute it  and sell  it,
-- alone or as a part of another product.
--
-- (C) Miguel Oliveira e Silva (mos@det.ua.pt), August-October 2000
--

deferred class THREAD

inherit
  THREAD_CONTROL

feature {THREAD}

  main is
      -- New thread starting point (main routine).
      -- Is not called directly, but throught 'start*' routines
      -- The new thread start object will be 'Current'
      -- The thread terminates at the end of 'main'
    deferred
    end;

feature

  start is
      -- start new thread
    require
      not is_expanded_type
    do
      c_start_pthread(Current,$c_main);
      from until
        pthread_running(pthread_data) or pthread_finish(pthread_data)
      loop end
    end;

  start_detached is
      -- start new thread on detached (unjoinable) state
    do
      c_start_detached_pthread(Current,$c_main);
      from until
        pthread_running(pthread_data) or pthread_finish(pthread_data)
      loop end
    end;

feature {NONE}

  frozen c_main(pth_data: POINTER) is
    do
      pthread_data := pth_data;
      set_pthread_running_state(pthread_data,true);
      main;
      set_pthread_running_state(pthread_data,false);
      set_pthread_finish_state(pthread_data,true);
    end;

  frozen c_start_pthread(c: like Current;rout: POINTER) is
    external "C"
    alias "start_pthread"
    end;

  frozen c_start_detached_pthread(c: like Current;rout: POINTER) is
    external "C"
    alias "start_detached_pthread"
    end;

  frozen pthread_finish(td: POINTER): BOOLEAN is
    external "C"
    end;

  frozen set_pthread_running_state(td: POINTER;st: BOOLEAN) is
    external "C"
    end;

  frozen set_pthread_finish_state(td: POINTER;st: BOOLEAN) is
    external "C"
    end;

end -- THREAD

