-- This  file  is   free  software,  which   comes  along  with   thread-safe
-- SmallEiffel. This software  is distributed  in the  hope that  it will  be
-- useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
-- you want, provided this  header is kept unaltered,  and a notification  of
-- the changes is  added. You  are allowed to  redistribute it  and sell  it,
-- alone or as a part of another product.
--
-- (C) Miguel Oliveira e Silva (mos@det.ua.pt), October-November 2000
--
-- This class allows the creation and the final execution of a
-- multiple thread synchronization structure.
--
-- This structure is defined using a generalized pipeline analogy,
-- in which the user can define any number of sequential "pipes",
-- each one with any number of concurrent threads. The synchronization
-- imposed be the pipeline structure is [off course] strictly observed
-- on pipeline execution.
--

class THREAD_PIPELINE

creation
  make

feature

  make is
    local
      pipe: LINKED_LIST[THREAD];
    do
      !!pipeline.make;
      !!pipe.make;
      pipeline.add_last(pipe);
    end;

  add_thread(thr: THREAD) is
      -- adds a new concurrent thread to current [last] "pipe".
    require
      thr /= Void
    do
      pipeline.last.add_last(thr)
    end;

  empty_pipe: BOOLEAN is
      -- is current pipe empty?
    do
      Result := pipeline.last.count = 0
    end;

  new_pipe is
      -- appends a new empty "pipe" to pipeline.
    require
      current_pipe_not_empty: not empty_pipe
    local
      pipe: LINKED_LIST[THREAD];
    do
      !!pipe.make;
      pipeline.add_last(pipe);
    end;

  start is
      -- starts pipeline thread execution.
      -- exits only on pipeline termination.
    local
      c,d: INTEGER;
      pipe: LINKED_LIST[THREAD];
    do
      from
        c := pipeline.lower
      until
        c > pipeline.upper
      loop
        pipe := pipeline @ c;
        from
          d := pipe.lower
        until
          d > pipe.upper
        loop
          (pipe @ d).start;
          d := d + 1
        end;
        from
          d := pipe.lower
        until
          d > pipe.upper
        loop
          (pipe @ d).join;
          d := d + 1
        end;
        c := c + 1
      end
    end;

feature {NONE}

  pipeline: LINKED_LIST[LINKED_LIST[THREAD]];

end -- THREAD_PIPELINE

