--          This file is part of SmallEiffel The GNU Eiffel Compiler.
--       Copyright (C) 1994-2002 LORIA - INRIA - U.H.P. Nancy 1 - FRANCE
--          Dominique COLNET and Suzanne COLLIN - SmallEiffel@loria.fr
--                       http://SmallEiffel.loria.fr
-- SmallEiffel is  free  software;  you can  redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free
-- Software  Foundation;  either  version  2, or (at your option)  any  later
-- version. SmallEiffel is distributed in the hope that it will be useful,but
-- WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-- or  FITNESS FOR A PARTICULAR PURPOSE.   See the GNU General Public License
-- for  more  details.  You  should  have  received a copy of the GNU General
-- Public  License  along  with  SmallEiffel;  see the file COPYING.  If not,
-- write to the  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-- Boston, MA 02111-1307, USA.
--
-- Changed for thread-safety, by Miguel Oliveira e Silva (mos@det.ua.pt)
--
deferred class ONCE_ROUTINE
   --
   -- Root of ONCE_FUNCTION or ONCE_PROCEDURE.
   --

inherit EFFECTIVE_ROUTINE;

feature {NONE}

   pretty_print_once_or_do is
      do
         fmt.put_string(fz_once);
      end;

feature -- pthread extension

   thread_scope: BOOLEAN; -- processor

   set_thread_scope is
      do
         thread_scope := true
      end;

   process_scope: BOOLEAN; -- all-processors

   set_process_scope is
      do
         process_scope := true
      end;

   object_scope: BOOLEAN;

   set_object_scope is
      do
         object_scope := true
      end;

   class_scope: BOOLEAN;

   set_class_scope is
      do
         class_scope := true
      end;

   class_thread_scope: BOOLEAN is
      do
         Result := class_scope and thread_scope
      end;

   class_process_scope: BOOLEAN is
      do
         Result := class_scope and process_scope
       end;

   object_thread_scope: BOOLEAN is
      do
         Result := object_scope and thread_scope
      end;

   object_process_scope: BOOLEAN is
      do
         Result := object_scope and process_scope
      end;

   key_scope: BOOLEAN is
      do
        Result := once_key_list /= Void and then not once_key_list.is_empty
      end;

   once_key_list: FIXED_ARRAY[STRING];

   set_once_keys(list: FIXED_ARRAY[STRING]) is
      require
         list /= Void
      do
        !!once_key_list.from_collection(list)
      end;

end -- ONCE_ROUTINE

