/* c_thread.h
** This  file  is   free  software,  which   comes  along  with   thread-safe
** SmallEiffel. This software  is distributed  in the  hope that  it will  be
** useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
** you want, provided this  header is kept unaltered,  and a notification  of
** the changes is  added. You  are allowed to  redistribute it  and sell  it,
** alone or as a part of another product.
**
** (C) Miguel Oliveira e Silva (mos@det.ua.pt), August-October 2000
**
** Support for multi-threaded SmallEiffel programs.
**
** Changed in August 2001
*/

#ifndef C_THREAD_H
#define C_THREAD_H

/* Assumes:
  T6 -> BOOLEAN - 1: TRUE, 0: FALSE
  T8 -> POINTER (void*)
*/

#include <pthread.h>
#include <errno.h>

#ifdef __USE_UNIX98
typedef pthread_rwlock_t rwlock_t;
#else
typedef struct
{
  pthread_mutex_t  mutex;
  pthread_cond_t   read;
  pthread_cond_t   write;
  int              r_active;
  int              w_active;
  int              r_wait;
  int              w_wait;
} rwlock_t;
#endif

extern pthread_key_t _thread_data_key_;
extern rwlock_t *_thread_creation_tree_rwl_;

typedef struct __Thread_Data__
{
  pthread_t id;

  T6 running;
  T6 finish;
  T6 detached;
  T6 error; /* POSIX thread library error */

    /* tree with thread creation structure */
  struct __Thread_Data__ *parent;
  struct __Thread_Data__ *first_child;
  struct __Thread_Data__ *sibling;

#ifndef _BOOST_MODE_
  /* SE global variables */

  struct _se_dump_stack* _se_dst;
  int _se_rspf;
  int _se_require_uppermost_flag;
  int _se_require_last_result;
#endif

  rwlock_t *rwl;
} _Thread_Data_;

#ifndef _BOOST_MODE_
#define se_dst ((_Thread_Data_*)pthread_getspecific(_thread_data_key_))->_se_dst
#define se_rspf ((_Thread_Data_*)pthread_getspecific(_thread_data_key_))->_se_rspf
#define se_require_uppermost_flag ((_Thread_Data_*)pthread_getspecific(_thread_data_key_))->_se_require_uppermost_flag
#define se_require_last_result ((_Thread_Data_*)pthread_getspecific(_thread_data_key_))->_se_require_last_result
#endif

void _add_new_thread_(_Thread_Data_ *td);

#endif

