/* once_thread.c
** This  file  is   free  software,  which   comes  along  with   thread-safe
** SmallEiffel. This software  is distributed  in the  hope that  it will  be
** useful, but WITHOUT  ANY WARRANTY; without  even the  implied warranty  of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. You can modify it  as
** you want, provided this  header is kept unaltered,  and a notification  of
** the changes is  added. You  are allowed to  redistribute it  and sell  it,
** alone or as a part of another product.
**
** (C) Miguel Oliveira e Silva (mos@det.ua.pt), October 2000
**
** Changed in August 2003
**
** Thread safe SmallEiffel
*/

/*
** thread once support routines
**
** select:
**   0 -> Static system thread once support
**   1 -> Dynamic object thread once support
*/

pthread_key_t _thread_once_key_[2];
pthread_once_t _thread_once_key_once_ = PTHREAD_ONCE_INIT;

#define INITIAL_THREAD_ONCE_ARRAY_SIZE 32
#define INCREMENT_THREAD_ONCE_ARRAY_SIZE 16

void _thread_once_key_once_routine_(void)
{
  if (pthread_key_create(&_thread_once_key_[0],NULL) ||
      pthread_key_create(&_thread_once_key_[1],NULL))
  {
    /* lacks appropriate error handling code !!! */
  }
}

void _check_alloc_thread_once_routine_call_(int select,int index)
{
  _Threaded_Once_Data_ *tod;
  int i;

  pthread_once(&_thread_once_key_once_,_thread_once_key_once_routine_);

  tod = (_Threaded_Once_Data_*)pthread_getspecific(_thread_once_key_[select]);

  if (tod == NULL)
  {
    tod = se_malloc(sizeof(_Threaded_Once_Data_));
    if (pthread_setspecific(_thread_once_key_[select],(void*)tod))
    {
      /* lacks appropriate error handling code !!! */
    }

    tod->size = INITIAL_THREAD_ONCE_ARRAY_SIZE;
    if (tod->size <= index)
      tod->size = index + 1;
    tod->called_array = se_malloc(tod->size * sizeof(void *));
    for(i = 0;i < tod->size;i++)
      tod->called_array[i] = NULL; /* "not called" initialization */
  }
  else if (tod->size <= index)
  {
    int old_size;

    old_size = tod->size;
    tod->size += INCREMENT_THREAD_ONCE_ARRAY_SIZE;
    if (tod->size <= index)
      tod->size = index + 1;
    tod->called_array = se_realloc(tod->called_array,tod->size * sizeof(void *));
    for(i = old_size;i < tod->size;i++)
      tod->called_array[i] = NULL; /* "not called" initialization */
  }
}

int _is_first_thread_once_routine_call_(int select,int index)
{
  _Threaded_Once_Data_ *tod;
  int result = 0;

  _check_alloc_thread_once_routine_call_(select,index);
  tod = (_Threaded_Once_Data_*)pthread_getspecific(_thread_once_key_[select]);
  if (tod->called_array[index] == NULL)
    result = 1;

  return result;
}

void _set_first_thread_once_routine_call_(int select,int index)
{
  _Threaded_Once_Data_ *tod;

  _check_alloc_thread_once_routine_call_(select,index);
  tod = (_Threaded_Once_Data_*)pthread_getspecific(_thread_once_key_[select]);
  if (tod->called_array[index] == NULL)
    tod->called_array[index] = (void*)1; /* called from now on */
}

void _set_result_of_thread_once_routine_(int select,int index,void *result)
{
  _Threaded_Once_Data_ *tod;

  _check_alloc_thread_once_routine_call_(select,index);
  tod = (_Threaded_Once_Data_*)pthread_getspecific(_thread_once_key_[select]);
  tod->called_array[index] = result;
}

void *_get_result_of_thread_once_routine_(int select,int index)
{
  _Threaded_Once_Data_ *tod;
  void *result = NULL;

  _check_alloc_thread_once_routine_call_(select,index);
  tod = (_Threaded_Once_Data_*)pthread_getspecific(_thread_once_key_[select]);
  if (tod != NULL && tod != (_Threaded_Once_Data_ *)1)
    result = tod->called_array[index];

  return result;
}

/*
** Extra dynamic object thread once support code
*/

pthread_mutex_t _mutex_object_thread_once_routines_ = PTHREAD_MUTEX_INITIALIZER;
static int _num_object_thread_once_routines_ = 0;

void _set_index_object_thread_once_routine(int *index)
{
  pthread_mutex_lock(&_mutex_object_thread_once_routines_);
  if ((*index) == -1)
  {
    (*index) = _num_object_thread_once_routines_;
    _num_object_thread_once_routines_++;
  }
  pthread_mutex_unlock(&_mutex_object_thread_once_routines_);
}

/*
** Once key support code
*/

#ifdef _TSSE_INCLUDE_ONCE_KEY_ROUTINES_

int equal_keys(char *k1,char *k2)
{
  int i;

  for(i = 0;k1[i] && k2[i] && k1[i] == k2[i];i++)
    ;

  return (k1[i] || k2[i]) ? 0 : 1;
}

void _reset_once_key_(T8 a)
{
  char *key = (char *)a;
  int i;

    /* no mutex needed here since this part of the array is immutable */
  for(i = 0;(i < _once_key_list_size_) &&
      equal_keys(key,_once_key_list_[i].key);i++)
    ;
  if(i < _once_key_list_size_)
  {
    pthread_mutex_lock(&_once_key_list_mutex_);
    _once_key_list_[i].count++;
    pthread_mutex_unlock(&_once_key_list_mutex_);
  }
  else
  {
    /* ignores non-existent keys */
  }
}

void _reset_all_once_keys_(void)
{
  int i;

  pthread_mutex_lock(&_once_key_list_mutex_);
  for(i = 0;i < _once_key_list_size_;i++)
    _once_key_list_[i].count++;
  pthread_mutex_unlock(&_once_key_list_mutex_);
}

int _is_first_once_key_routine_call_(char *key,int count)
{
  int i,result = 0;

  for(i = 0;(i < _once_key_list_size_) &&
      equal_keys(key,_once_key_list_[i].key);i++)
    ;
  pthread_mutex_lock(&_once_key_list_mutex_);
  result = (count == _once_key_list_[i].count) ? 0 : 1;
  pthread_mutex_unlock(&_once_key_list_mutex_);

  return result;
}

void _set_once_key_routine_called_(char *key,int *count)
{
  int i;

  for(i = 0;(i < _once_key_list_size_) &&
      equal_keys(key,_once_key_list_[i].key);i++)
    ;
  pthread_mutex_lock(&_once_key_list_mutex_);
  (*count) = _once_key_list_[i].count;
  pthread_mutex_unlock(&_once_key_list_mutex_);
}

#endif

